/*
 * Copy a file out of a filesystem.
 *
 * Copyright (C) 2019, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <string.h>

#include <sys/stat.h>

#include <ext2fs/ext2fs.h>

#include "file.h"
#include "image.h"
#include "path.h"
#include "session.h"



/* Copy buffer size. */

static int BUFSIZE = 4096;



/* Alternative metadata set by options. */

extern struct metadata md;



/* Copy a file out of the filesystem image. */

static int _copy_file_out(const char *target, const char *filename,
                          int target_is_file, ext2_filsys fs,
                          ext2_ino_t ino_file)
{
    int retval = 0;
    ext2_file_t file;

    /* Copying details. */

    FILE *fp;
    char buf[BUFSIZE];
    unsigned int got;
    size_t written;

    /* Open the file in the source directory. */

    if (ext2fs_file_open(fs, ino_file, 0, &file))
        return 1;

    /* Open a file in the target directory. */

    if (target_is_file)
        fp = fopen(target, "w");
    else
        fp = open_file_in_dir(target, path_basename(filename), "w");

    /* Copy the file content. */

    if (fp != NULL)
    {
        do
        {
            if (ext2fs_file_read(file, buf, BUFSIZE, &got))
            {
                retval = 1;
                goto close_files;
            }

            while (got)
            {
                written = fwrite(buf, sizeof(char), got, fp);
                got -= written;
            }

        } while (got);
    }

close_files:
    fclose(fp);
    ext2fs_file_close(file);

    return retval;
}



/* Make a new directory in the external environment. */

static int _make_directory(const char *target, const char *basename,
                           ext2_filsys fs, ext2_ino_t ino)
{
    errcode_t retval;
    struct ext2_inode inode;
    char target_path[strlen(target) + strlen(basename) + 2];

    sprintf(target_path, "%s/%s", target, basename);

    retval = ext2fs_read_inode(fs, ino, &inode);

    if (retval)
        return retval;

    return mkdir(target_path, inode.i_mode);
}

/* Copy a file from the filesystem into the external environment. */

static int _copy_out(ext2_filsys fs, const char *source, const char *target,
                     int target_is_file)
{
    ext2_ino_t ino_source;

    if (image_find_by_path(fs, source, &ino_source))
    {
        fprintf(stderr, "Failed to find file: %s\n", source);
        return 1;
    }

    /* Test whether the filename references a file. */

    if (image_isdir_by_path(fs, source))
    {
        if (_make_directory(target, path_basename(source), fs, ino_source))
        {
            fprintf(stderr, "Failed to make directory: %s/%s\n", target,
                    path_basename(source));
            return 1;
        }
    }
    else if (image_isfile_by_path(fs, source))
    {
        if (_copy_file_out(target, source, target_is_file, fs, ino_source))
        {
            fprintf(stderr, "Failed to read from file: %s\n", source);
            return 1;
        }
    }
    else
    {
        fprintf(stderr, "Object type not supported: %s\n", source);
        return 1;
    }

    /* NOTE: Overwrite/update metadata where appropriate. */

    return 0;
}

/* Copy source files out of the filesystem image into the external environment. */

int copy_out(ext2_filsys fs, int argc, char *argv[])
{
    int i;

    /* Target filename details. */

    char *target = argv[argc - 1];
    int target_is_file;

    /* Locate the target and test whether it is a directory. */

    if (!isdir(target))
    {
        /* Only a new or existing file in an existing directory is permitted. */

        if (isfile(target) || isdir_dirname(target))
            target_is_file = 1;
        else
        {
            fprintf(stderr, "Target is not a directory: %s\n", target);
            return 1;
        }

        /* Only permit a target file when one source file is given. */

        if (argc > 2)
        {
            fprintf(stderr, "Target can only be a file when copying a single file: %s\n", target);
            return 1;
        }
    }
    else
        target_is_file = 0;

    /* Copy each source file to the target directory. */

    for (i = 0; i < argc - 1; i++)
        if (_copy_out(fs, argv[i], target, target_is_file))
            return 1;

    return 0;
}

/* vim: tabstop=4 expandtab shiftwidth=4
*/
